from .entries import entry_types
from .nacha_file import NachaFile


class Parser(object):
    _blocking_file_control_record = '9' * 94

    def __init__(self, confirmation_file=False, rejection_file=False):
        self.confirmation_file = confirmation_file
        self.rejection_file = rejection_file

    def parse_lines(self, file_data):
        lines = file_data.split('\n')

        for line in lines:
            if line.strip() == Parser._blocking_file_control_record:
                # There shouldn't be any content after a blocking control record
                break

            code = line[0]
            entry = entry_types[code].from_text(line)
            yield entry

    def parse(self, file_data):
        """
        {'1': entries.FileHeader,
         '5': entries.CompanyBatchHeader,
         '6': entries.EntryDetail,
         '7': entries.EntryAddenda,
         '8': entries.CompanyBatchControl,
         '9': entries.FileControl}
        :param file_data:
        :return:
        """
        return self.lines_to_nacha_file(self.parse_lines(file_data))

    @staticmethod
    def lines_to_nacha_file(lines):
        nacha_file = None

        for line in lines:
            assert line.code in entry_types.keys()

            if line.code == '1':
                nacha_file = NachaFile(file_header=line)
            elif line.code == '5':
                nacha_file.add_batch(batch_header=line)
            elif line.code == '6':
                nacha_file \
                    .batches[-1].add_entry(entry_detail=line)
            elif line.code == '7':
                nacha_file \
                    .batches[-1] \
                    .entries[-1].add_addenda(entry_addenda=line)
            elif line.code == '8':
                nacha_file \
                    .batches[-1].batch_control = line
            elif line.code == '9':
                nacha_file.file_control = line

        return nacha_file


parser = Parser()
parse_lines = parser.parse_lines
parse = parser.parse



from .entries import EntryAddenda


class Addenda(object):
    def __init__(self, entry_addenda=None, **kwargs):
        self.entry_addenda = entry_addenda or EntryAddenda(**kwargs)

    @property
    def error_code(self):
        return self.entry_addenda.error_code

    @error_code.setter
    def error_code(self, value):
        self.set_error_code(value)

    def set_error_code(self, error_code):
        self.entry_addenda.error_code = error_code



import datetime

import math

from comanage_nacha.exceptions import EntryClosedError
from .batch import Batch
from .entries import FileHeader, FileControl


class NachaFile(object):
    def __init__(self, file_header=None, file_control=None, batches=None, **kwargs):
        kwargs.setdefault('file_creation_date', datetime.date.today())
        kwargs.setdefault('file_creation_time', datetime.datetime.utcnow().time())
        self.file_header = file_header or FileHeader(**kwargs)
        self.file_control = file_control
        self.batches = batches or []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def add_batch(self, **kwargs):
        if self.file_control is not None:
            raise EntryClosedError("This file has already been closed")
        kwargs.setdefault('error_code', self.file_header.error_code)
        batch = Batch(self.batch_count + 1, **kwargs)
        self.batches.append(batch)
        return batch

    @property
    def error_code(self):
        return self.file_header.error_code

    @error_code.setter
    def error_code(self, value):
        self.set_error_code(value)

    def set_error_code(self, error_code):
        self.file_header.error_code = error_code
        for batch in self.batches:
            batch.set_error_code(error_code)

    def calculate_entry_addenda_record_count(self):
        return sum(batch.entry_count + sum(entry.addenda_count for entry in batch.entries)
                   for batch in self.batches)

    def calculate_block_count(self):
        return int(math.ceil((
            1 +  # File Header
            len(self.batches) * 2 +  # Batch Headers and Controls
            self.calculate_entry_addenda_record_count() +
            1  # File Control
        ) / 10.0))

    @property
    def batch_count(self):
        return len(self.batches)

    def close(self):
        self.file_control = FileControl(
            batch_count=len(self.batches),
            block_count=self.calculate_block_count(),
            entry_addenda_record_count=self.calculate_entry_addenda_record_count(),
            entry_hash_total=str(sum(int(batch.entry_hash) for batch in self.batches))[-10:],
            total_file_debit_entry_amount=sum(batch.calculate_total_batch_debit_entry() for batch in self.batches),
            total_file_credit_entry_amount=sum(batch.calculate_total_batch_credit_entry() for batch in self.batches),
        )

    @property
    def lines(self):
        yield self.file_header
        for batch in self.batches:
            for line in batch.lines:
                yield line
        yield self.file_control

    def render_to_string(self):
        return '\n'.join(line.dumps() for line in self.lines)



import abc

import six

from comanage_nacha.error_codes import REJECTION_REASONS
from comanage_nacha.utils.entry_formatter import EntryFormatter


class EntryBase(six.with_metaclass(abc.ABCMeta)):
    code = '0'
    format = '0'
    error_code = None

    @abc.abstractmethod
    def loads(self, line):
        raise NotImplementedError

    def dumps(self):
        dumped = EntryFormatter().format(self.format, **self.__dict__)

        if self.rejected:
            return (
                dumped[:79] +
                'REJ0' +
                self.error_code +
                dumped[87:]
            )
        return dumped

    @classmethod
    def from_text(cls, text):
        instance = cls()
        instance.loads(text)
        return instance

    @property
    def rejected(self):
        return hasattr(self, 'error_code') and self.error_code is not None

    @property
    def error_reason(self):
        return hasattr(self, 'error_code') and REJECTION_REASONS.get(self.error_code, None)


__all__ = ['EntryBase']



from comanage_nacha.entries.entrybase import EntryBase


class EntryAddenda(EntryBase):
    code = '7'
    format = (
        '7'
        '{addenda_type_code!d:02d}'
        '{payment_information!s: >80s}'
        '{addenda_sequence_number!d:04d}'
        '{entry_detail_sequence_number!d:07d}'
    )

    def __init__(self,
                 addenda_type_code=5,
                 payment_information='',
                 addenda_sequence_number=1,
                 entry_detail_sequence_number=1,
                 error_code=None,
                 ):
        self.addenda_type_code = addenda_type_code
        self.payment_information = payment_information
        self.addenda_sequence_number = addenda_sequence_number
        self.entry_detail_sequence_number = entry_detail_sequence_number
        self.error_code = error_code

    def loads(self, line):
        self.addenda_type_code = int(line[1:1 + 2])
        self.payment_information = line[3:3 + 80].rstrip()
        self.addenda_sequence_number = int(line[83:83 + 4])
        self.entry_detail_sequence_number = int(line[87:87 + 7])

        if line[79:79 + 4] == 'REJ0':
            self.error_code = line[83:83 + 4]



import six


class EnumMeta(type):
    def __new__(cls, what, bases, dikt):
        for k, v in list(dikt.items()):
            dikt[v] = k
        return type(what, bases, dikt)(dikt)


Enum = six.with_metaclass(EnumMeta, dict)


class ConfirmationMessageCodes(Enum):
    NO_MESSAGE = ''
    FILE_RECEIVED = '01'
    FILE_INVALID_ENTRY_DATES = '02'
    FILE_CONTROL_TOTALS_INCORRECT = '03'
    FILE_FAILED_SECONDARY_EDIT = '04'
    FILE_SEVERE_ERRORS = '05'
    FILE_DUPLICATE_QUALIFIER = '06'
    FILE_NEW_QUAILIFIER_ASSIGNED = '07'
    FILE_TEST_NOT_PROCESSED = 'TT'


class ServiceClassCodes(Enum):
    MIXED_DEBITS_CREDITS = 200
    CREDITS = 220
    DEBITS = 225


class StandardEntryClasses(Enum):
    ARC = 'ARC'  # Accounts Receivable Entry
    CIE = 'CIE'  # Customer Initiated Entry
    MTE = 'MTE'  # Machine Transfer Entry
    PBR = 'PBR'  # Consumer Cross-Border Payment
    POP = 'POP'  # Point-of-Purchase
    PPD = 'PPD'  # Prearranged Payment & Deposit
    POS = 'POS'  # Point of Sale Entry/Shared Network Transaction
    SHR = 'SHR'  # Point of Sale Entry/Shared Network Transaction
    RCK = 'RCK'  # Re-presented Check Entry
    TEL = 'TEL'  # Telephone-Initiated Entry
    WEB = 'WEB'  # Internet-Initiated Entry
    CBR = 'CBR'  # Corporate Cross-Border Payment
    CCD = 'CCD'  # Cash Concentration or Disbursement
    CTX = 'CTX'  # Corporate Trade Exchange
    ACK = 'ACK'  # Acknowledgment Entries
    ATX = 'ATX'  # Acknowledgment Entries
    ADV = 'ADV'  # Automated Accounting Advice
    COR = 'COR'  # Automated Notification of Change or Refused Notification of Change
    DNE = 'DNE'  # Death Notification Entry
    ENR = 'ENR'  # Automated Enrollment Entry
    TRC = 'TRC'  # Truncated Entries
    TRX = 'TRX'  # Truncated Entries
    XCK = 'XCK'  # Destroyed Check Entry'


class TransactionCodes(Enum):
    # credit checking
    CHECKING_RETURNED_CREDIT = 21
    CHECKING_CREDIT = 22
    CHECKING_PRE_NOTE_CREDIT = 23

    # debit checking
    CHECKING_RETURNED_DEBIT = 26
    CHECKING_DEBIT = 27
    CHECKING_PRE_NOTE_DEBIT = 28

    # credit savings
    SAVINGS_RETURNED_CREDIT = 31
    SAVINGS_CREDIT = 32
    SAVINGS_PRE_NOTE_CREDIT = 33

    # debit savings
    SAVINGS_RETURNED_DEBIT = 36
    SAVINGS_DEBIT = 37
    SAVINGS_PRE_NOTE_DEBIT = 38



from .entrybase import EntryBase


class FileControl(EntryBase):
    code = '9'
    format = (
        '9'
        '{batch_count!d:06d}'
        '{block_count!d:06d}'
        '{entry_addenda_record_count!d:08d}'
        '{entry_hash_total!d:010d}'
        '{total_file_debit_entry_amount!d:012d}'
        '{total_file_credit_entry_amount!d:012d}'
        ' '
        '{message_code1!s: <2s}'
        '{message_code2!s: <2s}'
        '{message_code3!s: <2s}'
        '                                '
    )

    def __init__(self,
                 batch_count=None,
                 block_count=None,
                 entry_addenda_record_count=None,
                 entry_hash_total=None,
                 total_file_debit_entry_amount=None,
                 total_file_credit_entry_amount=None,
                 message_code1='',
                 message_code2='',
                 message_code3='',
                 ):
        self.batch_count = batch_count
        self.block_count = block_count
        self.entry_addenda_record_count = entry_addenda_record_count
        self.entry_hash_total = entry_hash_total
        self.total_file_debit_entry_amount = total_file_debit_entry_amount
        self.total_file_credit_entry_amount = total_file_credit_entry_amount
        self.message_code1 = message_code1
        self.message_code2 = message_code2
        self.message_code3 = message_code3

    def loads(self, line):
        self.batch_count = int(line[1:1 + 6])
        self.block_count = int(line[7:7 + 6])
        self.entry_addenda_record_count = int(line[13:13 + 8])
        self.entry_hash_total = int(line[21:21 + 10])
        self.total_file_debit_entry_amount = int(line[31:31 + 12])
        self.total_file_credit_entry_amount = int(line[43:43 + 12])

        self.message_code1 = line[56:56 + 2].strip()
        self.message_code2 = line[58:58 + 2].strip()
        self.message_code3 = line[60:60 + 2].strip()

    @property
    def message_codes(self):
        return [i for i in [self.message_code1, self.message_code2, self.message_code3] if i]



import datetime

from comanage_nacha.entries.entrybase import EntryBase


class FileHeader(EntryBase):
    code = '1'
    format = (
        "1"
        "{priority_code!d:02d}"
        "{wells_fargo_routing_number!s: >10s}"
        "{file_id!s:10s}"
        "{file_creation_date:%y%m%d}"
        "{file_creation_time:%H%M}"
        "{file_id_modifier!s:1s}"
        "{record_size!d:0>3d}"
        "{blocking_factor!d:0>2d}"
        "{format_code!d:1d}"
        "{origination_bank!s: <23s}"
        "{company_name!s: <23s}"
        "{reference_code!s: >8s}"
    )

    def __init__(self,
                 priority_code=1,
                 wells_fargo_routing_number='091000019',
                 file_id=None,
                 file_creation_date=None,
                 file_creation_time=None,
                 file_id_modifier='A',
                 record_size=94,
                 blocking_factor=10,
                 format_code=1,
                 origination_bank='WELLS FARGO',
                 company_name=None,
                 reference_code=None,
                 error_code=None,
                 ):
        self.priority_code = priority_code
        self.wells_fargo_routing_number = wells_fargo_routing_number
        self.file_id = file_id
        self.file_creation_date = file_creation_date
        self.file_creation_time = file_creation_time
        self.file_id_modifier = file_id_modifier
        self.record_size = record_size
        self.blocking_factor = blocking_factor
        self.format_code = format_code
        self.origination_bank = origination_bank
        self.company_name = company_name
        self.reference_code = reference_code
        self.error_code = error_code

    def loads(self, line):
        self.priority_code = int(line[1: 1 + 2])
        self.wells_fargo_routing_number = line[3: 3 + 10].strip()
        self.file_id = line[13: 13 + 10]
        self.file_creation_date = datetime.datetime.strptime(line[23: 23 + 6], '%y%m%d').date()
        self.file_creation_time = datetime.datetime.strptime(line[29: 29 + 4], '%H%M').time()
        self.file_id_modifier = line[33: 33 + 1]
        self.record_size = int(line[34: 34 + 3])
        self.blocking_factor = int(line[37: 37 + 2])
        self.format_code = int(line[39: 39 + 1])
        self.origination_bank = line[40: 40 + 23].rstrip()
        self.company_name = line[63: 63 + 23].rstrip()
        self.reference_code = line[86: 86 + 8].lstrip()
        if line[79:79 + 4] == 'REJ0':
            self.error_code = line[83: 83 + 4]



import random

from comanage_nacha import Parser
from ..error_codes import REJECTION_REASONS


def filter_rejection_lines(lines):
    """
    Filter lines to create a rejection only file.

    All parents or children of a rejected line are returned,

    :param lines:
    :return:
    """
    nacha = Parser.lines_to_nacha_file(lines)

    if not nacha.error_code:
        for batch in list(nacha.batches):
            if not batch.error_code:
                for entry in list(batch.entries):
                    if not entry.error_code:
                        for addenda in list(entry.addenda):
                            if not addenda.error_code:
                                entry.addenda.remove(addenda)
                        if not entry.addenda:
                            batch.entries.remove(entry)
                if not batch.entries:
                    nacha.batches.remove(batch)
        if not nacha.batches:
            return []

    return nacha.lines


def file_rejection(nacha, error_code=None):
    if not error_code:
        error_code = filter(lambda code: code[0] in '019', REJECTION_REASONS.keys())
    nacha.set_error_code(error_code)


def batch_rejection(nacha, error_code=None):
    if not error_code:
        error_code = filter(lambda code: code[0] in '58', REJECTION_REASONS.keys())
    random.choice(nacha.batches).set_error_code(error_code)


def entry_rejection(nacha, error_code=None):
    if not error_code:
        error_code = filter(lambda code: code[0] in '6', REJECTION_REASONS.keys())
    all_entries = [entry
                   for batch in nacha.batches
                   for entry in batch.entries]
    random.choice(all_entries).set_error_code(error_code)


def addenda_rejection(nacha, error_code=None):
    if not error_code:
        error_code = filter(lambda code: code[0] in '7', REJECTION_REASONS.keys())
    all_addenda = [addenda
                   for batch in nacha.batches
                   for entry in batch.entries
                   for addenda in entry.addenda]
    random.choice(all_addenda).set_error_code(error_code)


weighted_methods = sum(([method] * weight for weight, method in [
    (10, file_rejection),
    (20, batch_rejection),
    (40, entry_rejection),
]), [])


def random_reject(nacha):
    return random.choice(weighted_methods)(nacha)



import string


class EntryFormatter(string.Formatter):

    def format_field(self, value, format_spec):
        return super(EntryFormatter, self).format_field(value, format_spec)

    def convert_field(self, value, conversion):
        if conversion == 'd':
            return int(value)
        if value is None and conversion == 's':
            return ''
        return super(EntryFormatter, self).convert_field(value, conversion)
